/* chilnoise.c -- calculates noise parameters from Chilbolton-format
   radar data.

   Copyright (C) 2001 Robin Hogan
*/

#include <stdlib.h>
#include <string.h>
#include <math.h>
#include "chil.h"

#define MAX_ERR 0.5
/* Parameters in determining the ZDR offset using vertical data */
#define MIN_ZDR_ELEV 80
#define MAX_LDR_FOR_ZDR -15
#define THRESHOLD_ZDR 4


static
int
compare(const void *a, const void *b)
{
  return *(float *)a > *(float *)b ? 1 : -1;
}

static
float
median(float *x, int points)
{
  if (points < 2) {
    /* Only one point */
    return *x;
  }
  qsort(x, points, sizeof(float), compare);      
  if (points % 2) {
    /* Odd number of points */
    return x[points/2];
  }
  else {
    return 0.5 * (x[(points/2)-1] + x[points/2]);
  }
}


int
main(int argc, char **argv)
{
  /* The `ChilboltonData' structure contains the state information 
   * as well as the radar data itself. */

  float ZHarr[MAX_RASTERS], ZVarr[MAX_RASTERS], Xarr[MAX_RASTERS];
  float ZDRarr[MAX_RASTERS], PHI0arr[MAX_RASTERS];

  long int ZHn = 0, ZVn = 0, Xn = 0, ZDRn = 0, PHI0n = 0;

  ChilboltonData data;
  int status, iarg = 0;

  float min_height = 14; /* Height above which we calculate the noise. */
  char force_min_height = 0;

  uchar cloud_data = 0;

  /* Interpret the command-line arguments as processing directives.
   * Any unrecognised arguments are ignored. This function also opens the
   * first radar file and positions it ready to read the first ray. */

  status = CHILInit(argc, argv, &data);

  /* If fewer than 2 arguments (including the command itself), show
   * usage information */

  if (argc < 2)
    {
      CHILUsage(&data);
      /*      CHILOptions(&data);*/

      CHILPrintf("\n"
		 "Option:\n"
		 "  -minheight f  use data above this height to estimate noise parameters\n");
      CHILPrintf("\n"
		 "The arguments accepted by chil2nc are also accepted by this program,\n"
		 "but it is recommended that you don't use them.\n");

      exit(0);
    }

  /* If there are any additional command-line arguments provided only 
   * by this program then search the argument list for them here. */
  
  while (++iarg < argc) {
    if (strcmp(argv[iarg], "-minheight") == 0) {
      if (argc > iarg+1) {
	min_height = atof(argv[++iarg]);
	force_min_height = 1;
      }
      else {
	CHILPrintf("Error: minimum height not specified\n");
	exit(1);
      }
    }
  }

  /* If the CHILInit() function is able to open the first (or only)
   * radar file without any problems then it will return CHIL_OK. */

  if (status == CHIL_OK)
    {

      int file, raster, scan;

      if (!data.quiet) {
	CHILPrintf("Tape Raster Scan Z-pixels    ZH-channel     ZV-channel     Cross-polar  PHIDP-rays   PHIDP\n");
	CHILPrintf("------------------------------------------------------------------------------------------\n");
      }

      while (1)
	{
	  int failures = 0, maxfailures = 10;

	  /* The sumations of ZH, ZV and the cross-polar channel. `2'
             indicates the sum of the square, for the purposes of
             calculating the error */
	  double ZHsum = 0, ZH2sum = 0, ZVsum = 0, ZV2sum = 0, Xsum = 0, X2sum = 0;
	  double ZDRsum = 0, ZDR2sum = 0, PHIDPsum = 0;

	  float mh = min_height;

	  word ZH_index = CHILFindZH(&data.ray);
	  word ZDR_index = CHILFindField(&data.ray, ZDR);
	  word LDR_index = CHILFindField(&data.ray, LDR);
	  word PHIDP_index = CHILFindField(&data.ray, PHIDP);

	  int npixels = 0, nzdrpixels = 0, nphipixels = 0;

	  while ((status = CHILFetchRay(&data)) != CHIL_OK)
	    {
	      if (status != CHIL_ENDOFBLOCK) {
		break;
	      }
	      if ((++failures) == maxfailures)
		{
		  CHILPrintf("Aborting: %d failed calls to CHILFetchRay - empty scans?\n");
		  break;
		}
	    }
	  if (status != CHIL_OK)
	    {
	      break;
	    }
 	  /* We have read the first ray of a new block of data, which for 
           * scanning data is a single scan, and for cloud data is a
           * single day. */

	  file = data.raster_info.file;
	  raster = data.raster_info.raster;
	  scan = data.chil_file.iscan+1;

	  if (!force_min_height && data.raster_info.system_id == SINGAPORE) {
	    mh = 20; /* Tropical data! */
	  }

	  cloud_data = data.raster_info.cloud_data;

          do
            {
	      uint maxzdrgate = 0;

	      if (ZH_index >= 0) {
		/* In the data.ray structure we have our processed data */
		uint igate = 0;

		while (igate < data.ray.ngates && sin(data.ray.elevation * M_PI/180)
		       * (data.ray.range[igate] + 0.001 * data.ray.range_offset) < mh) {
		  ++igate;
		}
		maxzdrgate = igate;

		for (; igate < data.ray.ngates; igate++) {
		  float zh;
		  float range2 = 0;

		  if (data.ray.range_normalised) {
		    range2 = 20 * log10(data.ray.range[igate]);
		  }
		  zh = *(data.ray.data[ZH_index] + igate) - range2;

		  ZHsum += zh;
		  ZH2sum += zh * zh;
		  if (ZDR_index >= 0) {
		    float zv = zh - *(data.ray.data[ZDR_index] + igate);
		    ZVsum += zv;
		    ZV2sum += zv * zv;
		    if (LDR_index >= 0) {
		      float x = zh + *(data.ray.data[LDR_index] + igate);
		      Xsum += x;
		      X2sum += x * x;
		    }
		  }
		  ++npixels;
		}
	      }

	      if (PHIDP_index >= 0) {
		float *phidp = data.ray.data[PHIDP_index];
		uint igate;
		for (igate = 6; igate < data.ray.ngates; igate++) {
		  float mean = (phidp[igate-4] + phidp[igate-3] + phidp[igate-2]
				+ phidp[igate-1] + phidp[igate]) / 5.0;
		  float variance = (phidp[igate-4]*phidp[igate-4] + phidp[igate-3]*phidp[igate-3]
				    + phidp[igate-2]*phidp[igate-2] + phidp[igate-1]*phidp[igate-1]
				    + phidp[igate]*phidp[igate])/5.0 - mean*mean;
		  if (variance < 9) {
		    PHIDPsum += mean;
		    ++nphipixels;
		    break;
		  }
		}
	      }

	      /* Try to calibrate ZDR: 75 degrees lets 3.5% of the horizontal component in */
	      if (ZH_index >= 0 && ZDR_index >= 0 && LDR_index >= 0
		  && npixels > 8 && data.ray.elevation > MIN_ZDR_ELEV) {
		int igate;
		int start_gate = (int) (1.0 + (4.0-data.ray.range[0])
					/ (data.ray.range[1]-data.ray.range[0]));
		if (start_gate < 1) {
		  start_gate = 1;
		}
		/* First do a high sensitivity noise subtraction to get high sensitivity LDR */
		CHILSubtractNoisePolar(&data.ray, ZHsum/npixels, ZVsum/npixels, Xsum/npixels, 1.13);
		/* Use high sensitivity LDR to throw away ground clutter */
		CHILSuppressClutterWithLdr(&data.ray, MAX_LDR_FOR_ZDR);
		/* Do a low sensitivity processing to get high accuracy ZDR */
		CHILSubtractNoisePolar(&data.ray, ZHsum/npixels, ZVsum/npixels, Xsum/npixels, THRESHOLD_ZDR);
		/* Throw away remaining anomalous echos clouds */
		CHILDoubleCleanRay(&data.ray);
		for (igate = start_gate; igate < maxzdrgate; igate++) {
		  if (*(data.ray.data[ZDR_index] + igate) > chil_no_signal_value + 1.0) {
		    float zdr = *(data.ray.data[ZDR_index] + igate);
		    ZDRsum += zdr;
		    ZDR2sum += zdr * zdr;
		    ++nzdrpixels;
		  }
		}
	      }
            }
          while ((status = CHILFetchRay(&data)) == CHIL_OK);

          /* The CHILFetchRay() function has returned either 
           * CHIL_ENDOFBLOCK or CHIL_ENDOFDATA. */

	  if (!data.quiet) {
	    CHILPrintf("%04u %6u %4u", file, raster, scan);
	  }

	  if (npixels) {
	    float ZHnoise = ZHsum / npixels;
	    float ZHerr = sqrt(ZH2sum / npixels - ZHnoise * ZHnoise);

	    if (!data.quiet) {
	      CHILPrintf(" %8d", npixels);
	      CHILPrintf(" %+8.3f%5.3f", ZHnoise, ZHerr);
	    }

	    if (ZHerr < MAX_ERR && ZHn < MAX_RASTERS) {
	      ZHarr[ZHn++] = ZHnoise;
	    }

	    if (ZDR_index >= 0) {
	      float ZVnoise = ZVsum / npixels;
	      float ZVerr = sqrt(ZV2sum / npixels - ZVnoise * ZVnoise);
	      
	      if (!data.quiet) {
		CHILPrintf(" %+8.3f%5.3f", ZVnoise, ZVerr);
	      }

	      if (ZVerr < MAX_ERR && ZVn < MAX_RASTERS) {
		ZVarr[ZVn++] = ZVnoise;
	      }
	      
	      if (LDR_index >= 0) {
		float Xnoise = Xsum / npixels;
		float Xerr = sqrt(X2sum / npixels - Xnoise * Xnoise);

		if (!data.quiet) {
		  CHILPrintf(" %+8.3f%5.3f", Xnoise, Xerr);
		}

		if (Xerr < MAX_ERR && Xn < MAX_RASTERS) {
		  Xarr[Xn++] = Xnoise;
		}

	      }

	      if (nzdrpixels) {
		float ZDRmean = ZDRsum / nzdrpixels;
		float ZDRerr = sqrt(ZDR2sum / nzdrpixels - ZDRmean * ZDRmean);

		if (!data.quiet) {
		  CHILPrintf(" (ZDR offset %+.3f%5.3f from %d pixels)", ZDRmean, ZDRerr, nzdrpixels);
		}

		if (ZDRerr < MAX_ERR && ZDRn < MAX_RASTERS) {
		  ZDRarr[ZDRn++] = ZDRmean;
		}

	      }

	    }

	  }
	  else if (nphipixels) {
	    float PHIDPfloor = PHIDPsum / nphipixels;
	    PHI0arr[PHI0n++] = PHIDPfloor;
	    if (!data.quiet) {
	      CHILPrintf("                                                       %10d  %+.3f", nphipixels, PHIDPfloor);
	    }
	  }

	  if (!data.quiet) {
	    CHILPrintf("\n");
	  }

          /* If CHILFetchRay() returned CHIL_ENDOFDATA then we have no
           * more data left and can exit from the program. */

          if (status == CHIL_ENDOFDATA)
            {
              break;
	    }

          /* Otherwise subsequent calls to CHILFetchRay() will return
           * rays from the next block. */

	}

      /* Finally we deallocate any dynamically-allocated data and issue
       * any delayed warnings. */
      CHILEnd(&data);

    }
  else {

    /* The CHILInit() function has failed to read any data. It will
     * announce on standard error what the problem was, but we may 
     * wish to add any tips to the user here. */

    CHILPrintf("Run with no options for more information\n");
    exit(1);
  }


  if (!cloud_data && (ZHn || PHI0n)) {
    float ZHmedian = 0.0, ZVmedian = 0.0, Xmedian = 0.0;

    if (!data.quiet) {
      CHILPrintf("--------------------------------------------------------------------------------------------\n");
      CHILPrintf("Recommended options:  ");
    }

    if (ZHn) {
      ZHmedian = median(ZHarr, ZHn);
    }

    if (ZVn) {
      ZVmedian = median(ZVarr, ZVn);
    }

    if (Xn) {
      Xmedian = median(Xarr, Xn);
    }

    if (ZVn) {
      printf("-polar %+.2f %+.2f %+.2f ", ZHmedian, ZVmedian, Xmedian);
    }
    else if (ZHn) {
      printf("-noise %+.2f ",  ZHmedian);
    }

    if (ZDRn) {
      printf("-ZDRcal %+.2f ", -median(ZDRarr, ZDRn));
    }

    if (PHI0n) {
      printf("-PHIDPcal %+.2f ", -median(PHI0arr, PHI0n));
    }

  }
  fflush(stdout);
  CHILPrintf("\n");

  /* Program completed successfully. */

  exit(0);
}
